# import suds.transport.https
# import suds.transport.http
#
# #!/usr/bin/python
#
# # urllib2 GSS-API (SPNEGO) using PyKerberos
#
# # Licensed under the Apache License, Version 2.0 (the "License");
# # you may not use this file except in compliance with the License.
# # You may obtain a copy of the License at
# #     http://www.apache.org/licenses/LICENSE-2.0
# # Unless required by applicable law or agreed to in writing, software
# # distributed under the License is distributed on an "AS IS" BASIS,
# # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# # See the License for the specific language governing permissions and
# # limitations under the License.
#
# import re
# import logging
# import kerberos
# from urllib2 import BaseHandler
# from urllib2 import HTTPPasswordMgr
#
# def getLogger():
#     log = logging.getLogger("http_kerberos_auth_handler")
#     handler = logging.StreamHandler()
#     formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
#     handler.setFormatter(formatter)
#     log.addHandler(handler)
#     return log
#
# log = getLogger()
# log.setLevel(logging.ERROR)
#
#
# class BaseKerberosHandler(BaseHandler):
#     """auth handler for urllib2 that does Kerberos HTTP Negotiate
#     Authentication
#     """
#
#     def negotiate_value(self, headers):
#         """checks for "Negotiate" in proper auth header
#         """
#         authreq = headers.get(self.auth_header, None)
#         log.debug('authreq = {0}'.format(authreq))
#
#         if authreq:
#             rx = re.compile(r'(?:.*,)*\s*Negotiate\s*([^,]*),?', re.I)
#             mo = rx.search(authreq)
#             if mo:
#                 return mo.group(1)
#             else:
#                 log.debug("regex failed on: %s" % authreq)
#
#         else:
#             log.debug("%s header not found" % self.auth_header)
#
#         return None
#
#     def __init__(self, password_mgr=None):
#         if password_mgr is None:
#             password_mgr = HTTPPasswordMgr()
#         self.retried = 0
#         self.context = None
#         self.password = password_mgr
#         self.add_password = self.password.add_password
#
#     def generate_request_header(self, req, headers, neg_value):
#         self.retried += 1
#         log.debug("retry count: %d" % self.retried)
#
#         host = req.get_host()
#         log.debug("req.get_host() returned %s" % host)
#
#         # Build the service name
#         if self.service is None:
#             self.service = "HTTP@%s" % host.rsplit(':', 1)[0]
#
#         # User credentials are in the Password Manager
#         log.debug("Finding credentials for url: %s" % req.get_full_url())
#         user, password = self.password.find_user_password(None, req.get_full_url())
#
#         # TODO: Enusre a TGT for the user principal exists, if not, use the password
#         # TODO: to aquire one.
#         # We currently rely on Nico's pykerberos patch to do this, one day it might be
#         # nice to do this using the open source kerbeos gss api 'C' libraries
#
#
#         # An empty username will attempt the current user
#         log.debug("Intialising Kerberos Context, service: %s UPN: %s" % (self.service, user))
#         result, self.context = kerberos.authGSSClientInit(self.service, user, password=password)
#
#         if result < 0:
#             log.warning("authGSSClientInit returned result %d" % result)
#             return None
#
#         log.debug("authGSSClientInit() succeeded")
#
#         result = kerberos.authGSSClientStep(self.context, neg_value)
#
#         if result < 0:
#             log.warning("authGSSClientStep returned result %d" % result)
#             return None
#
#         log.debug("authGSSClientStep() succeeded")
#
#         response = kerberos.authGSSClientResponse(self.context)
#         log.debug("authGSSClientResponse() succeeded")
#
#         return "Negotiate %s" % response
#
#     def authenticate_server(self, headers):
#         neg_value = self.negotiate_value(headers)
#         if neg_value is None:
#             log.critical("mutual auth failed. No negotiate header")
#             return None
#
#         result = kerberos.authGSSClientStep(self.context, neg_value)
#
#         if result < 1:
#             # this is a critical security warning
#             # should change to a raise --Tim
#             log.critical(
#                 "mutual auth failed: authGSSClientStep returned result %d" %
#                 result)
#
#     def clean_context(self):
#         if self.context is not None:
#             log.debug("cleaning context")
#             kerberos.authGSSClientClean(self.context)
#             self.context = None
#
#     def http_error_auth_reqed(self, host, req, headers):
#         neg_value = self.negotiate_value(headers)  # Check for auth_header
#         if neg_value is not None:
#             if not self.retried > 0:
#                 return self.retry_http_kerberos_auth(req, headers, neg_value)
#             else:
#                 return None
#         else:
#             self.retried = 0
#
#     def https_request(self, req):
#         neg_hdr = self.generate_request_header(req, None, '')
#         req.add_unredirected_header(self.authz_header, neg_hdr)
#         return req
#
#     def retry_http_kerberos_auth(self, req, headers, neg_value):
#         try:
#             neg_hdr = self.generate_request_header(req, headers, neg_value)
#
#             if neg_hdr is None:
#                 log.debug("neg_hdr was None")
#                 return None
#
#             req.add_unredirected_header(self.authz_header, neg_hdr)
#             resp = self.parent.open(req)
#
#             if resp.getcode() != 200:
#                 self.authenticate_server(resp.info())
#
#             return resp
#
#         except kerberos.GSSError, e:
#             self.clean_context()
#             self.retried = 0
#             log.critical("GSSAPI Error: %s/%s" % (e[0][0], e[1][0]))
#             return None
#
#         self.clean_context()
#         self.retried = 0
#
#
# class ProxyKerberosAuthHandler(BaseKerberosHandler):
#     """Kerberos Negotiation handler for HTTP proxy auth
#     """
#
#     authz_header = 'Proxy-Authorization'
#     auth_header = 'proxy-authenticate'
#
#     handler_order = 480  # before Digest auth
#
#     def http_error_407(self, req, fp, code, msg, headers):
#         log.debug("inside http_error_407")
#         host = req.get_host()
#         retry = self.http_error_auth_reqed(host, req, headers)
#         self.retried = 0
#         return retry
#
#
# class HTTPKerberosAuthHandler(BaseKerberosHandler):
#     """Kerberos Negotiation handler for HTTP auth
#     """
#
#     authz_header = 'Authorization'
#     auth_header = 'www-authenticate'
#
#     handler_order = 480  # before Digest auth
#
#     def __init__(self, service, keytab, pm):
#         # TODO: If the service is not set seach for it. If we still not found
#         # then we will try the standard service HTTP@fqdn
#         BaseKerberosHandler.__init__(self, pm)
#         self.service = service
#         self.keytab = keytab
#
#     def http_error_401(self, req, fp, code, msg, headers):
#         log.debug("inside http_error_401")
#         host = req.get_host()
#         retry = self.http_error_auth_reqed(host, req, headers)
#         self.retried = 0
#         return retry
#
#
# class KerberosHttpAuthenticated(suds.transport.https.HttpAuthenticated):
#     """
#     Provides Kerberos http authentication.
# 	    - service = kerberos service name (HTTP@hostname)
# 		- keytab
#     """
#
#     def __init__(self, **kwargs):
#         self.service = kwargs.get('service', None)
#         self.keytab = kwargs.get('keytab',None)
#         suds.transport.https.HttpAuthenticated.__init__(self, **kwargs)
#
#     def u2handlers(self):
#         handlers = suds.transport.http.HttpTransport.u2handlers(self)
#         handlers.append(HTTPKerberosAuthHandler(self.service, self.keytab, self.pm))
#         return handlers


# (c) 2015, Ian Clegg <ian.clegg@sourcewarp.com>
#
# winrmlib is licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__author__ = 'ian.clegg@sourcewarp.com'
import traceback
import re
import logging
import xmltodict
from requests import Session

from winrmlib.api.exception import WSManException
from winrmlib.api.exception import WSManOperationException
from winrmlib.api.exception import WSManAuthenticationException
from winrmlib.api.authentication import HttpCredSSPAuth
from winrmlib.api.authentication import HttpNtlmAuth

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

class Service(object):
    """
    SOAP Service
    """

    def __init__(self, endpoint, username, password, delegation=False, **kwargs):
        """
        # Kerberos authentication does not require a password if the MIT kerberos GSS-API mechanism already has a
        # valid service ticket for the WSMAN service on the target server. However if the callee wishes to use
        # credential delegation with Kerberos they must still provide a password (see Microsoft [MS-CSSP] Protocol)
        #
        # Credential delegation requires an authentication mechanism which provide message integrity and confidentiality
        # such as NTLM or Kerberos; basic authentication cannot be used.
        #
        # Kerberos requires the username in UPN (RFC xxxx) form. UPN or NetBIOS usernames can be used whith NTLM
        #
        # when TCP connection is closed, its auto-opened
        #
        """
        self.session = kwargs.get('session', Session())
        self.endpoint = endpoint
        self.session.auth = Service._determine_auth_mechanism(username, password, delegation)

    def invoke(self, headers, body):
        """
        Invokes the soap service
        """
        xml = Service._create_request(headers, body)

        try:
            response = self.session.post(self.endpoint, verify=False, data=xml)
            logging.debug(response.content)
        except Exception as e:
            traceback.print_exc()
            raise WSManException(e)

        if response.status_code == 200:
            return Service._parse_response(response.content)

        if response.status_code == 401:
            raise WSManAuthenticationException('the remote host rejected authentication')

        raise WSManException('the remote host returned an unexpected http status code: %s' % response.status_code)

    @staticmethod
    def _determine_auth_mechanism(username, password, delegation):
        """
        if the username contains at '@' sign we will use kerberos
        if the username contains a '/ we will use ntlm
        either NTLM or Kerberos. In fact its basically always Negotiate.
        """
        if re.match('(.*)@(.+)', username) is not None:
            if delegation is True:
                raise Exception('Kerberos is not yet supported, specify the username in <domain>\<username> form for NTLM')
            else:
                raise Exception('Kerberos is not yet supported, specify the username in <domain>>\<username> form for NTLM')

        # check for NT format 'domain\username' a blank domain or username is invalid
        legacy = re.match('(.*)\\\\(.*)', username)
        if legacy is not None:
            if not legacy.group(1):
                raise Exception('Please specify the Windows domain for user in <domain>\<username> format')
            if not legacy.group(2):
                raise Exception('Please specify the Username of the user in <domain>\<username> format')
            if delegation is True:
                return HttpCredSSPAuth(legacy.group(1), legacy.group(2), password)
            else:
                return HttpNtlmAuth(legacy.group(1), legacy.group(2), password)

        #return HttpCredSSPAuth("SERVER2012", "Administrator", password)
        # attempt NTLM (local account, not domain) - if username is '' then we try anonymous NTLM auth
        # as if anyone will configure that - uf!
        return HttpNtlmAuth('', username, password)

    @staticmethod
    def _create_request(headers, body):
        """
        Create the SOAP 1.2 Envelope
        An ordered dictionary is required to ensure the same order is reflected in the XML, otherwise the
        SOAP Body element would appear before the Header element.
        """
        envelope = OrderedDict()
        for (namespace, alias) in Service.Namespaces.items():
            envelope['@xmlns:' + alias] = namespace
        envelope['soap:Header'] = headers
        envelope['soap:Body'] = body
        return xmltodict.unparse({'soap:Envelope': envelope}, encoding='utf-8')

    @staticmethod
    def _parse_response(xml):
        """
        Attempt to parse the SOAP response and return a python object
        Raise a WSManException if a Fault is found
        """
        try:
            soap_response = xmltodict.parse(xml, process_namespaces=True, namespaces=Service.Namespaces)
        except Exception:
            logging.debug('unable to parse the xml response: %s', xml)
            raise WSManException("the remote host returned an invalid soap response")

        # the delete response has an empty body
        body = soap_response['soap:Envelope']['soap:Body']
        if body is not None and 'soap:Fault' in body:
            raise WSManOperationException(body['soap:Fault']['soap:Reason']['soap:Text']['#text'])
        return body


Service.Avaliable_Mechanisms = ["basic", "ntlm", "kerberos"]
Service.Namespaces = {
    'http://www.w3.org/2003/05/soap-envelope': 'soap',
    'http://schemas.xmlsoap.org/ws/2004/08/addressing': 'a',
    'http://schemas.dmtf.org/wbem/wsman/1/cimbinding.xsd': 'b',
    'http://schemas.xmlsoap.org/ws/2004/09/enumeration': 'n',
    'http://schemas.xmlsoap.org/ws/2004/09/transfer': 'x',
    'http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd': 'w',
    'http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd': 'p',
    'http://schemas.microsoft.com/wbem/wsman/1/windows/shell': 'rsp',
    'http://schemas.microsoft.com/wbem/wsman/1/config': 'cfg'
}



# (c) 2015, Ian Clegg <ian.clegg@sourcewarp.com>
#
# winrmlib is licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__title__ = 'winrmlib'
__author__ = 'ian.clegg@sourcewarp.com'
__license__ = 'Apache 2.0'


class AsHex:
    def __init__(self, s):
        self.s = s

    def __str__(self):
        return ":".join([hex(ord(c))[2:].zfill(2) for c in self.s])



import os
import versioneer
from setuptools import setup

project_name = 'winrmlib'

# versioneer configuration
versioneer.VCS = 'git'
versioneer.versionfile_source = os.path.join('winrmlib', '_version.py')
versioneer.versionfile_build = os.path.join('winrmlib', '_version.py')
versioneer.tag_prefix = ''
versioneer.parentdir_prefix = 'winrmlib'

# PyPi supports only reStructuredText, so pandoc should be installed
# before uploading package
try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except ImportError:
    long_description = ''

requires = [
    "pyopenssl >= 0.14",
    "requests == 2.5.0",
    "ntlmlib >= 0.69",
    "ordereddict",
    "xmltodict"
]

setup(
    name=project_name,
    version=versioneer.get_version(),
    description='Python library for Windows Remote Management with CredSSP and NTLMv2',
    long_description=long_description,
    keywords='winrm winrw ws-man ws-management powershell'.split(' '),
    author='Ian Clegg',
    author_email='ian.clegg@sourcewarp.com',
    url='https://github.com/ianclegg/winrmlib/',
    license='MIT license',
    packages=['winrmlib', 'winrmlib.api'],
    package_data={'winrmlib.api': ['assets/xsd/*.xsd', 'assets/*.wsdl']},
    install_requires=requires,
    cmdclass=versioneer.get_cmdclass(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Clustering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Systems Administration',
    ],
)
